@@ -1,14 +1,14 @@
 import copy
 import itertools
 import logging
-import spacy
 import warnings
+from types import ModuleType
+from typing import Callable, List, Optional, Tuple
+
 from functional import seq
 from spacy.matcher import Matcher
 from spacy.tokens import Span
 from spacy.tokens.underscore import get_ext_args
-from types import ModuleType
-from typing import Callable, List, Optional, Tuple
 
 from replacy import default_match_hooks
 from replacy.db import get_forms_lookup, get_match_dict, load_lm
@@ -24,7 +24,6 @@ from replacy.util import (
     make_doc_if_not_doc,
     set_known_extensions,
     validate_match_dict,
-    spacy_version,
 )
 from replacy.version import __version__
 
@@ -124,6 +123,7 @@ class ReplaceMatcher:
             self.match_dict = attach_debug_hook(self.match_dict)
         self.allow_multiple_whitespaces = allow_multiple_whitespaces
         self.matcher = Matcher(self.nlp.vocab)
+        self.predicates = {}
         self._init_matcher()
         self.spans: List[Span] = []
         self.max_suggestions_count = max_suggestions_count
@@ -151,40 +151,21 @@ class ReplaceMatcher:
         for match_name, ps in self.match_dict.items():
             patterns = copy.deepcopy(ps["patterns"])
 
-            if spacy_version() >= 3:
-                patterns = self._allow_multiple_whitespaces3(patterns)
-                patterns = self._remove_unsupported3(patterns)
-            else:
-                patterns = self._allow_multiple_whitespaces(patterns)
-                patterns = self._remove_unsupported(patterns)
+            patterns = self._allow_multiple_whitespaces(patterns)
+            patterns = self._remove_unsupported(patterns)
 
             match_hooks = ps.get("match_hook", [])
-            callback = self._get_callback(match_name, match_hooks)
-            self._add_matcher_rule(match_name, patterns, callback)
+            self.predicates[match_name] = get_predicates(
+                match_hooks, self.default_match_hooks, self.custom_match_hooks
+            )
+            self.matcher.add(match_name, patterns)
 
-    def _add_matcher_rule(self, match_name, patterns, callback):
-        if spacy_version() >= 3:
-            self.matcher.add(match_name, patterns, on_match=callback, greedy="LONGEST")
-        else:
-            self.matcher.add(match_name, callback, patterns)
+    @staticmethod
+    def _fix_alignment_multiple_whitespaces(alignments):
+        return [int(a / 2) for a in alignments]
 
-    def _allow_multiple_whitespaces(self, patterns):
-        """
-        allow matching tokens separated by multiple whitespaces
-        they may appear after normalizing nonstandard whitespaces
-        ex. "Here␣is␣a\u180E\u200Bproblem." -> "Here␣is␣a␣␣problem."
-        pattern can be preceded and followed by whitespace tokens
-        to keep preceded_by... with and succeeded_by... with match hooks working
-        """
-        if self.allow_multiple_whitespaces:
-            white_pattern = {"IS_SPACE": True, "OP": "?"}
-            normalized_patterns = [white_pattern]
-            for p in patterns:
-                normalized_patterns += [p, white_pattern]
-            patterns = normalized_patterns
-        return patterns
-
-    def _allow_multiple_whitespaces3(self, patterns):
+    @staticmethod
+    def _allow_multiple_whitespaces(patterns):
         """
         allow matching tokens separated by multiple whitespaces
         they may appear after normalizing nonstandard whitespaces
@@ -203,14 +184,8 @@ class ReplaceMatcher:
             patterns = normalized_patterns
         return patterns
 
-    def _remove_unsupported(self, patterns):
-        # remove custom attributes not supported by spaCy Matcher
-        for p in patterns:
-            if "TEMPLATE_ID" in p:
-                del p["TEMPLATE_ID"]
-        return patterns
-
-    def _remove_unsupported3(self, patterns):
+    @staticmethod
+    def _remove_unsupported(patterns):
         # remove custom attributes not supported by spaCy Matcher
         for pattern in patterns:
             for p in pattern:
@@ -218,49 +193,40 @@ class ReplaceMatcher:
                     del p["TEMPLATE_ID"]
         return patterns
 
-    def _get_callback(self, match_name, match_hooks):
-        """
-        Most matches have the same logic to be executed each time a match is found
-        Some matches have extra logic, defined in match_hooks
-        """
-        # Get predicates once, callback is returned in a closure with this information
-        predicates = get_predicates(
-            match_hooks, self.default_match_hooks, self.custom_match_hooks
-        )
+    def _callback(self, doc, match):
+        match_id, start, end, alignments = match
+        alignments = ReplaceMatcher._fix_alignment_multiple_whitespaces(alignments)
 
-        def cb(matcher, doc, i, matches):
-            match_id, start, end = matches[i]
+        match_name = self.nlp.vocab[match_id].text
 
-            for pred in predicates:
-                try:
-                    if pred(doc, start, end):
-                        return None
-                except IndexError:
-                    break
-            match_name = self.nlp.vocab[match_id].text
-            span = self.Span(doc, start, end)
+        for pred in self.predicates[match_name]:
+            try:
+                if pred(doc, start, end):
+                    return None
+            except IndexError:
+                break
 
-            # find in match_dict if needed
-            span._.match_name = match_name
+        span = self.Span(doc, start, end)
 
-            pre_suggestions = self.match_dict[match_name]["suggestions"]
+        # find in match_dict if needed
+        span._.match_name = match_name
 
-            span._.suggestions = []
+        pre_suggestions = self.match_dict[match_name]["suggestions"]
 
-            for i, x in enumerate(pre_suggestions):
-                span._.suggestions += self.process_suggestions(
-                    x, doc, start, end, match_name, i
-                )
+        span._.suggestions = []
 
-            for novel_prop, default_value in self.novel_prop_defaults.items():
-                setattr(
-                    span._,
-                    novel_prop,
-                    self.match_dict[match_name].get(novel_prop, default_value),
-                )
-            self.spans.append(span)
+        for i, x in enumerate(pre_suggestions):
+            span._.suggestions += self.process_suggestions(
+                x, doc, start, end, match_name, i, alignments
+            )
 
-        return cb
+        for novel_prop, default_value in self.novel_prop_defaults.items():
+            setattr(
+                span._,
+                novel_prop,
+                self.match_dict[match_name].get(novel_prop, default_value),
+            )
+        self.spans.append(span)
 
     def _set_scorer(self, lm_path):
         # The following is not ideal
@@ -307,13 +273,13 @@ class ReplaceMatcher:
         return spans
 
     def process_suggestions(
-        self, pre_suggestion, doc, start, end, match_name, pre_suggestion_id
+        self, pre_suggestion, doc, start, end, match_name, pre_suggestion_id, alignments
     ):
         # get token <-> pattern correspondence
         pattern = self.match_dict[match_name]["patterns"]
 
         suggestion_variants = self.suggestion_gen(
-            pre_suggestion, doc, start, end, pattern, pre_suggestion_id
+            pre_suggestion, doc, start, end, pattern, pre_suggestion_id, alignments
         )
         # assert there aren't more than max_suggestions_count
         # otherwise raise warning and return []
@@ -412,7 +378,14 @@ class ReplaceMatcher:
         self.spans = []
         doc = make_doc_if_not_doc(sent, self.nlp)
         # this fills up self.spans
-        self.matcher(doc)
+        matches = self.matcher(doc, with_alignments=True)
+
+        # do the callback here instead of to pass it as callback on match
+        # here we alignment information to use for pattern ref
+        # we don't have this info on match callback
+        for match in matches:
+            self._callback(doc, match)
+
         for _, component in self.pipeline:
             # the default pipeline will:
             # sort suggestions by lm score
