@@ -199,31 +199,3 @@ class Inflector:
             return self.inflect_token(word, tag=tag, pos=pos)
         elif isinstance(word, str):
             return self.inflect_string(word, tag=tag, pos=pos)
-
-    def insert(self, doc, suggestion: str, index: int):
-        """
-        Returns the sentence with inserted inflected token.
-        If inflection is not supported - returns the original sentence.
-        ex. She washed her eggs. -> She ate her eggs.
-        If many inflections returned, take the first form.
-        """
-
-        # if string passed, conversion to doc
-        try:
-            doc.text
-        except AttributeError:
-            doc = self.nlp(doc)
-
-        infl_tokens = self.auto_inflect(doc, suggestion, index)
-
-        if len(infl_tokens):
-            infl_token = infl_tokens[0]
-
-        if infl_token:
-            token = doc[index]
-            changed_sent = "".join(
-                [doc.text[: token.idx], infl_token, doc.text[token.idx + len(token) :],]
-            )
-            return changed_sent
-        else:
-            return doc.text
